/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.fukurou.util;

import static org.opengion.fukurou.util.HybsConst.CR;				// 6.1.0.0 (2014/12/26) refactoring

import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.BufferedInputStream;

import java.util.List;
import java.util.ArrayList;

import org.apache.poi.hssf.record.Record;
import org.apache.poi.hssf.record.CellRecord;
// import org.apache.poi.hssf.record.RowRecord;
import org.apache.poi.hssf.record.SSTRecord;
import org.apache.poi.hssf.record.BOFRecord;
import org.apache.poi.hssf.record.EOFRecord;
import org.apache.poi.hssf.record.BoundSheetRecord;
import org.apache.poi.hssf.record.LabelSSTRecord;
import org.apache.poi.hssf.record.NumberRecord;
import org.apache.poi.hssf.record.BoolErrRecord;
import org.apache.poi.hssf.record.FormulaRecord;
import org.apache.poi.hssf.record.StringRecord;
import org.apache.poi.hssf.eventusermodel.HSSFEventFactory;
import org.apache.poi.hssf.eventusermodel.HSSFListener;
import org.apache.poi.hssf.eventusermodel.HSSFRequest;

import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.ErrorConstants;
import org.apache.poi.ss.util.NumberToTextConverter;
import org.apache.poi.poifs.filesystem.POIFSFileSystem;

/**
 * POI による、Excel(xls)の読み取りクラスです。
 *
 * xls形式のEXCELを、イベント方式でテキストデータを読み取ります。
 * このクラスでは、HSSF(.xls)形式のファイルを、ExcelReaderEvent を介したイベントで読み取ります。
 * ExcelReaderEventイベントは、openGion形式のファイル読み取りに準拠した方法をサポートします。
 * ※ openGion形式のEXCELファイルとは、#NAME 列に、カラム名があり、#で始まる
 *    レコードは、コメントとして判断し、読み飛ばす処理の事です。
 *
 * @og.rev 6.0.3.0 (2014/11/13) 新規作成
 * @og.group ファイル入力
 *
 * @version  6.0
 * @author   Kazuhiko Hasegawa
 * @since    JDK7.0,
 */
public final class ExcelReader_HSSF {
	/** このプログラムのVERSION文字列を設定します。	{@value} */
	private static final String VERSION = "6.0.3.0 (2014/11/13)" ;

//	private static final String CR = System.getProperty("line.separator");

	/**
	 * 引数ファイル(Excel)を、HSSFイベントモデルを使用してテキスト化します。
	 *
	 * ExcelReaderEventは、EXCEL読み取り処理用の統一されたイベント処理クラスです。
	 * openGion特有のEXCEL処理方法(#NAME , 先頭行#コメントなど)を実装しています。
	 * これは、HSSFやXSSFの処理を、統一的なイベントモデルで扱うためです。
	 * SSモデルが良いのですが、巨大なXSSF(.xlsx)ファイルを解析すると、OutOfMemoryエラーが
	 * 発生する為、個々に処理する必要があります。
	 * あくまで、読み取り限定であれば、こちらのイベントモデルで十分です。
	 *
	 * @og.rev 6.0.3.0 (2014/11/13) 新規作成
	 *
	 * @param	fname 入力ファイル名
	 * @param	event イベント処理するオブジェクト
	 */
	public static void excelReader( final String fname , final ExcelReaderEvent event ) {
		InputStream fin = null;
		InputStream din = null;

		try {
			fin = new BufferedInputStream( new FileInputStream( fname ) );
			final POIFSFileSystem poifs = new POIFSFileSystem( fin );

			din = poifs.createDocumentInputStream( "Workbook" );

			final HSSFRequest req = new HSSFRequest();
			req.addListenerForAllRecords( new ExcelListener( event ) );
			final HSSFEventFactory factory = new HSSFEventFactory();

			factory.processEvents( req, din );
		}
		catch( IOException ex ) {
			final String errMsg = "ファイルの読取処理に失敗しました。"
								+ " filename=" + fname + CR
								+ ex.getMessage() ;
			throw new RuntimeException( errMsg , ex );
		}
		finally {
			Closer.ioClose( din );
			Closer.ioClose( fin );
		}
	}

	/**
	 * HSSF(.xls)処理に特化したイベント処理を行う、HSSFListener の実装内部クラス。
	 *
	 * HSSFListener のイベント処理を、ExcelReaderEvent に変換します。
	 * これは、HSSFやXSSFの処理を、統一的なイベントモデルで扱うためです。
	 * SSモデルが良いのですが、巨大なXSSF(.xlsx)ファイルを解析すると、OutOfMemoryエラーが
	 * 発生する為、個々に処理する必要があります。
	 * あくまで、読み取り限定であれば、こちらのイベントモデルで十分です。
	 *
	 * 読み書きも含めた EXCEL処理を行うには、ExcelModel クラスが別にあります。
	 *
	 * @og.rev 6.0.3.0 (2014/11/13) 新規作成
	 */
	private static final class ExcelListener implements HSSFListener {
		private final ExcelReaderEvent	event;

		private final List<String> shtNms;		// シート名の一括登録
//		private List<String> shtNms = new ArrayList<String>();	// シート名の一括登録
		private SSTRecord sstrec;				// LabelSSTRecord のインデックスに対応した文字列配列

		private int		shtNo			= -1;	// 最初に見つけたときに、＋＋ するので初期値は －１にしておく
		private String	shtNm			;		// BOFRecord でキャッシュしておきます。
		private boolean isNextRecord	;		// FormulaRecord で、次のレコードに値があるかどうかの判定
		private boolean isReadSheet		= true;	// シートの読み取りを行うかどうか

		private int		rcdLvl	;				// BOFRecord で＋１、EOFRecord で－１ して、シートの EOFRecord の判定に使う。

		private int		rowNo	;				// 処理中の行番号(0～)
		private int		colNo	;				// 処理中の列番号(0～)

		private boolean	isDebug	;				// デバッグフラグ

		/**
		 * ExcelReaderEvent を引数に取るコンストラクタ
		 *
		 * HSSFListener のイベント処理を、ExcelReaderEvent に変換します。
		 *
		 * @og.rev 6.0.3.0 (2014/11/13) 新規作成
		 *
		 * @param	event イベント処理するオブジェクト
		 */
		public ExcelListener( final ExcelReaderEvent event ) {
			this.event = event ;
			shtNms     = new ArrayList<String>();	// シート名の一括登録
		}

		/**
		 * HSSFListener のイベントを受け取るメソッド
		 *
		 * @og.rev 6.1.0.0 (2014/12/26) シートの数のイベント
		 *
		 * @param record	イベント時に設定されるレコード
		 * @see	org.apache.poi.hssf.eventusermodel.HSSFListener
		 */
		@Override
		public void processRecord( final Record record ) {
			if( record instanceof CellRecord ) {
				final CellRecord crec = (CellRecord)record;
				rowNo = crec.getRow() ;
				if( event.isSkip( rowNo ) ) { return; }		// 行のスキップ判定
				colNo = crec.getColumn();
			}

			String val = null;
			switch( record.getSid() ) {
				// the BOFRecord can represent either the beginning of a sheet or the workbook
				case BOFRecord.sid:				// Beginning Of File 
					final BOFRecord bof = (BOFRecord)record;
					if( bof.getType() == BOFRecord.TYPE_WORKSHEET ) {
						// 6.1.0.0 (2014/12/26) シートの数のイベント
						// シート一覧の読み取り後、最初のレコードの判定に、shtNo を使います。
						if( shtNo < 0 ) { event.sheetSize( shtNms.size() ); }

						shtNo++ ;						// 現在のシート番号。初期値が、-1 してあるので、先に ＋＋ する。
						shtNm = shtNms.get( shtNo ) ;	// 現在のシート名。
						rcdLvl = 0;						// シートの開始
						isReadSheet = event.startSheet( shtNm,shtNo );
						if( isDebug ) { System.out.println( "① BOFRecord:" + record ); }
					}
					else {
						rcdLvl++;						// シート以外の開始
					}
					break;
				case EOFRecord.sid:				// End Of File record
					if( rcdLvl == 0 ) {					// シートの終了
						event.endSheet( shtNo );
						isReadSheet = true;
						if( isDebug ) { System.out.println( "② EOFRecord" + record ); }
					}
					else {
						rcdLvl--;						// シート以外の終了
					}
					break;
				case BoundSheetRecord.sid:		// シート一覧(一括で最初にイベントが発生する)
					if( isDebug ) { System.out.println( "③ BoundSheetRecord" ); }
					final BoundSheetRecord bsr = (BoundSheetRecord) record;
					shtNms.add( bsr.getSheetname() );
					break;
				case SSTRecord.sid:				// Static String Table Record
					if( isDebug ) { System.out.println( "④ SSTRecord" ); }
					sstrec = (SSTRecord)record;		// LabelSSTRecord のインデックスに対応した文字列配列
		//			for (int k = 0; k < sstrec.getNumUniqueStrings(); k++) {
		//				System.out.println("table[" + k + "]=" + sstrec.getString(k));
		//			}
					break;
		//		case RowRecord.sid:				// stores the row information for the sheet
		//			if( isDebug ) { System.out.println( "⑤ RowRecord" ); }
		//			RowRecord rowrec = (RowRecord) record;
		//			System.out.println("Row=[" + rowrec.getRowNumber() + "],Col=["
		//					+ rowrec.getFirstCol() + "]-[" + rowrec.getLastCol() + "]" );
		//			break;
				case NumberRecord.sid:			// extend CellRecord
					if( isReadSheet ) {
						final NumberRecord numrec = (NumberRecord)record;
						val = NumberToTextConverter.toText(numrec.getValue());
					}
					break;
				// SSTRecords store a array of unique strings used in Excel.
				case LabelSSTRecord.sid:		// extend CellRecord
					if( isReadSheet ) {
						final LabelSSTRecord lrec = (LabelSSTRecord)record;
						val = sstrec.getString(lrec.getSSTIndex()).getString();
					}
					break;
				case BoolErrRecord.sid:			// extend CellRecord
					if( isReadSheet ) {
						final BoolErrRecord berec = (BoolErrRecord)record;
						final byte errVal = berec.getErrorValue();
						val = errVal == 0 ? Boolean.toString( berec.getBooleanValue() )
										  : ErrorConstants.getText( errVal );
					}
					break;
				case FormulaRecord.sid:			// extend CellRecord
					if( isReadSheet ) {
						final FormulaRecord frec = (FormulaRecord)record;
						switch (frec.getCachedResultType()) {
							case Cell.CELL_TYPE_NUMERIC:
								final double num = frec.getValue();
								if( Double.isNaN(num) ) {
									// Formula result is a string
									// This is stored in the next record
									isNextRecord = true;
								}
								else {
									val = NumberToTextConverter.toText( num );
								}
								break;
							case Cell.CELL_TYPE_BOOLEAN:
								val = Boolean.toString(frec.getCachedBooleanValue());
								break;
							case Cell.CELL_TYPE_ERROR:
								val = ErrorConstants.getText(frec.getCachedErrorValue());
								break;
							case Cell.CELL_TYPE_STRING:
								isNextRecord = true;
								break;
							default : break;
						}
					}
					break;
				case StringRecord.sid:			// FormulaRecord の場合の次のレコードに値が設定されている
					if( isReadSheet ) {
						if( isNextRecord ) {
							// String for formula
							final StringRecord srec = (StringRecord)record;
							val = srec.getString();
							isNextRecord = false;
						}
					}
					break;
				default :
					break ;
			}
			if( val != null ) {
				//           値   行(Row) 列(Col)
				event.value( val, rowNo,  colNo );			// イベント処理
			}
		}
	}

	/**
	 * アプリケーションのサンプルです。
	 *
	 * 入力ファイル名 は必須で、第一引数固定です。
	 *
	 * Usage: java org.opengion.fukurou.util.ExcelReader_HSSF 入力ファイル名
	 *
	 * @og.rev 6.0.3.0 (2014/11/13) 新規作成
	 *
	 * @param	args	コマンド引数配列
	 */
	public static void main( final String[] args ) {
		final String usageMsg = "Usage: java org.opengion.fukurou.util.ExcelReader_HSSF 入力ファイル名" ;
		if( args.length == 0 ) {
			System.err.println( usageMsg );
			return ;
		}

		final String fname = args[0];

		ExcelReader_HSSF.excelReader(
			fname,
			new ExcelReaderEvent() {
				/**
				 * シートの読み取り開始時にイベントが発生します。
				 *
				 * @param   shtNm  シート名
				 * @param   shtNo  シート番号(0～)
				 * @return  true:シートの読み取り処理を継続します/false:このシートは読み取りません。
				 */
				public boolean startSheet( final String shtNm,final int shtNo ) {
					System.out.println( "S[" + shtNo + "]=" + shtNm );
					return super.startSheet( shtNm,shtNo );
				}

		//		public void columnNames( final String[] names ) {
		//			System.out.println( "NM=" + java.util.Arrays.toString( names ) );
		//		}

		//		public void values( final String[] vals,final int rowNo ) {
		//			System.out.println( "V[" + rowNo + "]=" + java.util.Arrays.toString( vals ) );
		//		}

		//		public boolean isSkip( final int rowNo ) {
		//			super.isSkip( rowNo );
		//			return false;
		//		}

				/**
				 * 読み取り状態の時に、rowNo,colNo にあるセルの値を引数にイベントが発生します。
				 *
				 * @param   val     文字列値
				 * @param   rowNo   行番号(0～)
				 * @param   colNo   列番号(0～)
				 * @return  読み取りするかどうか(true:読み取りする/false:読み取りしない)
				 */
				public boolean value( final String val,final int rowNo,final int colNo ) {
					System.out.println( "R[" + rowNo + "],C[" + colNo + "]=" + val );
					return super.value( val,rowNo,colNo );
				}
			}
		);
	}
}
